/* Interbench by Con Kolivas <kernel@kolivas.org> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sched.h>
#include <time.h>
#include <errno.h>
#include <semaphore.h>
#include <pthread.h>
#include <math.h>
//#include <sys/time.h>
//#include <sys/stat.h>

struct timespec myts;
unsigned long loops_per_ms = 0;
unsigned long ram = 0;
int duration = 25;

#define SEMS	4
struct sems {
	sem_t ready;
	sem_t start;
	sem_t stop;
	sem_t complete;
	sem_t *semlist[SEMS];
};

#define MAX_SAMPLES	10000
struct thread {
	void *name;
	char *label;
	pthread_t pthread;
	struct sems sem;
	unsigned long total_latency, max_latency, nr_samples, deadlines_met,
		missed_latency, missed_deadlines;
	unsigned long sample[MAX_SAMPLES];
};
struct thread audiothread, videothread, xthread, gamethread, burnthread,
	writethread, readthread;
#define BENCH_THREADS	4
#define THREADS		7
struct thread *threadlist[THREADS];

unsigned long latency_table[BENCH_THREADS][THREADS][5];

#define MIN_BLK_SIZE	1024

void terminal_error(char *name)
{
	perror(name);
	exit (1);
}

void terminal_fileopen_error(FILE *fp, char *name)
{
	perror(name);
	if (fclose(fp) == -1)
		terminal_error("fclose");
	exit (1);
}

unsigned long long get_nsecs(void)
{
	if (clock_gettime(CLOCK_MONOTONIC, &myts))
		terminal_error("clock_gettime");
	return (myts.tv_sec * 1000000000 + myts.tv_nsec );
}

unsigned long get_usecs(void)
{
	if (clock_gettime(CLOCK_MONOTONIC, &myts))
		terminal_error("clock_gettime");
	return (myts.tv_sec * 1000000 + myts.tv_nsec / 1000 );
}

static void burn_loops(unsigned long loops)
{
	unsigned long i;

	for (i = 0 ; i < loops ; i++);
}

/* Use this many usecs of cpu time */
static void burn_usecs(unsigned long usecs)
{
	unsigned long ms_loops;

	ms_loops = loops_per_ms / 1000 * usecs;
	burn_loops(ms_loops);
}

void periodic_schedule(struct thread *th, unsigned long run_usecs,
	unsigned long interval_usecs, unsigned long *deadline)
{
	unsigned long current_time, latency = 0;

	current_time = get_usecs();
	/* First sample latency is ignored */
	if (!th->nr_samples)
		current_time = *deadline;
	if (current_time > *deadline)
		latency = current_time - *deadline;
	th->sample[th->nr_samples] = latency;
	th->total_latency += latency;
	latency += th->missed_latency;
	if (latency > th->max_latency)
		th->max_latency = latency;

	burn_usecs(run_usecs);
	current_time = get_usecs();
	/*
	 * If we meet the deadline we move the deadline forward, otherwise
	 * we consider it a missed deadline and dropped frame etc.
	 */
	if (*deadline + interval_usecs > current_time) {
		usleep(*deadline + interval_usecs - current_time);
		th->deadlines_met++;
		*deadline += interval_usecs;
		th->total_latency += th->missed_latency;
		th->missed_latency = 0;
	} else {
		/*
		 * Missed deadline is counted but only used to calculate total
		 * deadlines possible under no load.
		 */
		th->missed_deadlines++;
		th->missed_latency = interval_usecs;
		*deadline += interval_usecs * 2;
	}
}

void initialise_thread_data(struct thread *th)
{
	th->max_latency = th->total_latency = th->deadlines_met =
		th->missed_latency = th->missed_deadlines = th->nr_samples = 0;
}

#define audio_interval	(50000)
#define audio_run	(audio_interval / 20)
/* We emulate audio by using 5% cpu and waking every 50ms */
void *emulate_audio(void)
{
	unsigned long deadline;
	struct thread *th = &audiothread;
	struct sems *s = &th->sem;

	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs();
	initialise_thread_data(th);

	while (1) {
		periodic_schedule(th, audio_run, audio_interval, &deadline);
		if (!sem_trywait(&s->stop))
			goto out;
		th->nr_samples++;
		if (th->nr_samples >= MAX_SAMPLES) {
			sem_wait(&s->stop);
			goto out;
		}
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* We emulate video by using 50% cpu and waking for 60fps */
#define video_interval	(1000000 / 60)
#define video_run	(video_interval / 2)
void *emulate_video(void)
{
	unsigned long deadline;
	struct thread *th = &videothread;
	struct sems *s = &th->sem;

	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs();
	initialise_thread_data(th);

	while (1) {
		periodic_schedule(th, video_run, video_interval, &deadline);
		if (!sem_trywait(&s->stop))
			goto out;
		th->nr_samples++;
		if (th->nr_samples >= MAX_SAMPLES) {
			sem_wait(&s->stop);
			goto out;
		}
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/*
 * We emulate X by running for a variable percentage of cpu from 0-100% 
 * in 1ms chunks.
 */
void *emulate_x(void)
{
	unsigned long deadline;
	struct thread *th = &xthread;
	struct sems *s = &th->sem;

	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs();
	initialise_thread_data(th);

	while (1) {
		int i, j;
		for (i = 0 ; i <= 100 ; i++) {
			j = 100 - i;
			periodic_schedule(th, i * 1000, j * 1000, 
				&deadline);
			deadline += i * 1000;
			if (!sem_trywait(&s->stop))
				goto out;
			th->nr_samples++;
			if (th->nr_samples >= MAX_SAMPLES) {
				sem_wait(&s->stop);
				goto out;
			}
		}
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* 
 * We emulate gaming by using 100% cpu and seeing how many frames (jobs
 * completed) we can do in that time. Deadlines are meaningless with 
 * unlocked frame rates
 */
#define game_interval	(1000)
#define game_run	(game_interval)
void *emulate_game(void)
{
	unsigned long deadline;
	struct thread *th = &gamethread;
	struct sems *s = &th->sem;

	sem_post(&s->ready);
	sem_wait(&s->start);
	deadline = get_usecs();
	initialise_thread_data(th);

	while (1) {
		periodic_schedule(th, game_run, game_interval, &deadline);
		if (!sem_trywait(&s->stop))
			goto out;
		th->nr_samples++;
		if (th->nr_samples >= MAX_SAMPLES) {
			sem_wait(&s->stop);
			goto out;
		}
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* Just burn cpu continuously */
void *emulate_burn(void)
{
	struct thread *th = &burnthread;
	struct sems *s = &th->sem;

	sem_post(&s->ready);
	sem_wait(&s->start);
	while (1) {
		burn_loops(loops_per_ms);
		if (!sem_trywait(&s->stop))
			goto out;
	}

out:
	sem_post(&s->complete);
	return NULL;
}

/* Write a file the size of ram continuously */
void *emulate_write(void)
{
	struct thread *th = &writethread;
	struct sems *s = &th->sem;
	FILE *fp;
	char *name = "interbench.write";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem;

	if (!(fp = fopen(name, "w")))
		terminal_error("fopen");
	if (stat(name, &statbuf) == -1)
		terminal_fileopen_error(fp, "stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	mem = ram / (statbuf.st_blksize / 1024);	/* kilobytes to blocks */
	if (!(buf = calloc(1, statbuf.st_blksize)))
		terminal_fileopen_error(fp, "calloc");
	if (fclose(fp) == -1)
		terminal_error("fclose");
	sync();
	sem_post(&s->ready);
	sem_wait(&s->start);

	while (1) {
		unsigned int i;

		if (!(fp = fopen(name, "w")))
			terminal_error("fopen");
		if (stat(name, &statbuf) == -1)
			terminal_fileopen_error(fp, "stat");
		for (i = 0 ; i < mem; i++) {
			if (fwrite(buf, statbuf.st_blksize, 1, fp) != 1)
				terminal_fileopen_error(fp, "fwrite");
			if (!sem_trywait(&s->stop))
				goto out;
		}
		if (fclose(fp) == -1)
			terminal_error("fclose");
	}

out:
	if (fclose(fp) == -1)
		terminal_error("fclose");
	if (remove(name) == -1)
		terminal_error("remove");
	sync();
	sem_post(&s->complete);
	return NULL;
}

void create_read_file(void)
{
	unsigned int i;
	FILE *fp;
	char *name = "interbench.read";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem;

	if (remove(name) == -1) {
		if (errno != ENOENT)
			terminal_error("remove");
	}
	if (!(fp = fopen(name, "w")))
		terminal_error("fopen");
	if (stat(name, &statbuf) == -1)
		terminal_fileopen_error(fp, "stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	if (!(buf = calloc(1, statbuf.st_blksize)))
		terminal_fileopen_error(fp, "calloc");
	mem = ram / (statbuf.st_blksize / 1024);	/* kilobytes to blocks */

	for (i = 0 ; i < mem; i++) {
		if (fwrite(buf, statbuf.st_blksize, 1, fp) != 1)
			terminal_fileopen_error(fp, "fwrite");
	}
	if (fclose(fp) == -1)
		terminal_error("fclose");
	sync();
}

/* Read a file the size of ram continuously */
void *emulate_read(void)
{
	struct thread *th = &readthread;
	struct sems *s = &th->sem;
	char *name = "interbench.read";
	void *buf = NULL;
	struct stat statbuf;
	unsigned long mem, bsize;
	int tmp;

	if ((tmp = open(name, O_RDONLY)) == -1) {
		if (errno == ENOENT) {
			create_read_file();
			if ((tmp = open(name, O_RDONLY)) == -1)
				terminal_error("open");
		} else
			terminal_error("open");
	}
	if (stat(name, &statbuf) == -1) 
		terminal_error("stat");
	if (statbuf.st_blksize < MIN_BLK_SIZE)
		statbuf.st_blksize = MIN_BLK_SIZE;
	bsize = statbuf.st_blksize;
	if (statbuf.st_size / 1024 / bsize != ram / bsize)
		create_read_file();
	mem = ram / (bsize / 1024);	/* kilobytes to blocks */
	if (!(buf = malloc(bsize)))
		terminal_error("malloc");

	sem_post(&s->ready);
	sem_wait(&s->start);

	while (1) {
		int rd;

		/* 
		 * We have to read the whole file before quitting the load
		 * to prevent the data being cached for the next read. This
		 * is also the reason the file is the size of physical ram.
		 */
		while ((rd = read(tmp , buf, bsize)) > 0);
		if(!sem_trywait(&s->stop))
			goto out;
		if (lseek(tmp, (off_t)0, SEEK_SET) == -1)
			terminal_error("lseek");
	}

out:
	sem_post(&s->complete);
	return NULL;
}

void calibrate_loop(void)
{
	unsigned long long start_time, loops_per_msec, run_time = 0;
	unsigned long loops;

	loops_per_msec = 100000;
redo:
	while (run_time / 1000 != 1000) {
		loops = loops_per_msec;
		start_time = get_nsecs();
		burn_loops(loops);
		run_time = get_nsecs() - start_time;
		loops_per_msec = (1000000 * loops_per_msec / run_time ? :
			loops_per_msec);
	}

	/* Rechecking after a pause increases reproducibility */
	sleep(1);
	loops = loops_per_msec;
	start_time = get_nsecs();
	burn_loops(loops);
	run_time = get_nsecs() - start_time;
	if (run_time / 1000 != 1000)
		goto redo;

	loops_per_ms = loops_per_msec;
}

/*
 * Signal the threads to start. Calling these before stop_* will make the
 * thread restart.
 */
void start_thread(struct thread *th)
{
	sem_wait(&th->sem.ready);
	sem_post(&th->sem.start);
}

void stop_thread(struct thread *th)
{
	sem_post(&th->sem.stop);
	sem_wait(&th->sem.complete);

	/* Kill the thread */
	if (pthread_join(th->pthread, NULL))
		terminal_error("pthread_join");
}

void initialise_thread(struct thread *th)
{
	int i;
	/* Initialise the semaphores */
	for (i = 0 ; i < SEMS ; i++) {
		if (sem_init(th->sem.semlist[i], 1, 0))
			terminal_error("sem_init");
	}
	/* Create the threads */
	if (pthread_create(&th->pthread, NULL, th->name, (void *)0))
		terminal_error("pthread_create");
}

/* Format output to appropriate amount of significant digits and 9 columns */
char *formatted_float(float f)
{
	char *s;

	if (f >= 100)
		sprintf(s, "%.0f", f);
	else if (f >= 10)
		sprintf(s, "%.1f", f);
	else
		sprintf(s, "%.2f", f);
	return s;
}

void show_latencies(int i)
{
	struct thread *th = threadlist[i];
	double average_latency, max_latency, deadlines_met, sd;
	unsigned long nr_samples, noload_deadlines, s;
	long double variance = 0;
	char *output;

	/* 
	 * Total deadlines possible is calculated by deadlines met +
	 * deadlines missed under no load
	 */
	noload_deadlines = (latency_table[i][i][3] +
		latency_table[i][i][4]) ? : 1;
	nr_samples = th->nr_samples;
	/* 
	 * Landing on the boundary of a deadline can make loaded runs appear
	 * to do more work than unloaded.
	 */
	if (nr_samples > latency_table[i][i][2])
		nr_samples = latency_table[i][i][2] ? : 1;
	average_latency = th->total_latency / nr_samples;
	average_latency /= 1000;
	max_latency = th->max_latency;
	max_latency /= 1000;
	deadlines_met = (double)th->deadlines_met / noload_deadlines * 100;
	if (deadlines_met >= 99)
		deadlines_met = 100;

	for (s = 0 ; s < nr_samples; s++) {
		long double diff = (long double)average_latency -
			(long double)th->sample[s];
		variance += diff * diff;
	}
	variance /= nr_samples - 1 ? : 1;
	sd = sqrtl(variance) / 1000;

	output = formatted_float(average_latency);
	printf("%.*s", (9 - sizeof(output)), output);
	output = formatted_float(sd);
	printf(" +/- %.*s", (9 - sizeof(output)), output);

#if 0	/* max latency is often bogus because it is often one dropped frame wide */
	if (max_latency >= 100)
		printf("%9.0f\t\t\t", max_latency);
	else
		printf("%7.1f\t\t\t", max_latency);
#endif

	printf("\t%9lu", nr_samples);
	output = formatted_float(deadlines_met);
	printf("\t\t%.*s\n", (9 - sizeof(output)), output);
}

void fill_table(int i, int j)
{
	struct thread *th = threadlist[i];

	latency_table[i][j][0] = th->total_latency;
	latency_table[i][j][1] = th->max_latency;
	latency_table[i][j][2] = th->nr_samples;
	latency_table[i][j][3] = th->deadlines_met;
	latency_table[i][j][4] = th->missed_deadlines;
}

void usage(void)
{
	printf("interbench v0.08 by Con Kolivas\n");
	printf("interbench [-l <int>] [-u] [-t <int] [-b] [-h]\n");
	printf(" -l\tUse known loops per sec (default is to benchmark)\n");
	printf(" -u\tImitate uniprocessor\n");
	printf(" -t\tSeconds to run each benchmark (default 25)\n");
	printf(" -b\tBenchmark loops_per_ms even if it is already known\n");
	printf(" -h\tShow this help\n");
}

void setup_threads(void)
{
	int i;

	threadlist[0] = &audiothread;
	threadlist[1] = &videothread;
	threadlist[2] = &xthread;
	threadlist[3] = &gamethread;
	threadlist[4] = &burnthread;
	threadlist[5] = &writethread;
	threadlist[6] = &readthread;

	audiothread.name = *emulate_audio;
	videothread.name = *emulate_video;
	xthread.name = *emulate_x;
	gamethread.name = *emulate_game;
	burnthread.name = *emulate_burn;
	writethread.name = *emulate_write;
	readthread.name = *emulate_read;

	audiothread.label = "Audio";
	videothread.label = "Video";
	xthread.label = "X";
	gamethread.label = "Gaming";
	burnthread.label = "Burn";
	writethread.label = "Write";
	readthread.label = "Read";

	for (i = 0 ; i < THREADS ; i++) {
		struct thread *th = threadlist[i];
		th->sem.semlist[0] = &th->sem.ready;
		th->sem.semlist[1] = &th->sem.start;
		th->sem.semlist[2] = &th->sem.stop;
		th->sem.semlist[3] = &th->sem.complete;
	}
}

void get_ram(void)
{
	FILE *meminfo;

	if(!(meminfo = fopen("/proc/meminfo", "r")))
		terminal_error("fopen");
	fscanf(meminfo, "MemTotal: %lu kB", &ram);
	fclose(meminfo);
}

void get_fifo(void)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = 99;
	if (sched_setscheduler(0, SCHED_FIFO, &sp) == -1) {
		if (errno != EPERM)
			terminal_error("sched_setscheduler");
		fprintf(stderr, "could not set RT scheduling\n");
	}
}

void get_normal(void)
{
	struct sched_param sp;
	memset(&sp, 0, sizeof(sp));
	sp.sched_priority = 0;
	if (sched_setscheduler(0, SCHED_OTHER, &sp) == -1) {
		fprintf(stderr, "could not unset RT scheduling!\n");
	}
}

void bench(struct thread *th)
{
	get_fifo();
	start_thread(th);
	sleep(duration);
	stop_thread(th);
	get_normal();
}

int main(int argc, char **argv)
{
	int q, i, j, affinity, benchmark = 0;
	FILE *fp;
	/* 
	 *This file stores the loops_per_ms to be reused in a filename that
	 * can't be confused
	 */
	char *fname = "interbench.loops_per_ms";

	while ((q = getopt(argc, argv, "hl:ut:b")) != -1) {
		switch (q) {
			case 'h':
				usage();
				return (0);
			case 'l':
				loops_per_ms = atoi(optarg);
				break;
			case 'u':
				affinity = 1;
				break;
			case 't':
				duration = atoi(optarg);
				break;
			case 'b':
				benchmark = 1;
				break;
			default:
				usage();
				return (1);
		}
	}

	if (affinity) {
#ifdef CPU_SET	/* Recent glibc expects cpu_set_t */
		cpu_set_t cpumask;

		CPU_ZERO(&cpumask);
		CPU_SET(0, &cpumask);
#else		/* Old glibc expects unsigned long */
		unsigned long cpumask = 1;
#endif
		if (sched_setaffinity(0, sizeof(cpumask), &cpumask) == -1) {
			if (errno != EPERM)
				terminal_error("sched_setaffinity");
			fprintf(stderr, "could not set cpu affinity\n");
		}
	}

	if (benchmark)
		loops_per_ms = 0;
	/* 
	 * Try to get loops_per_ms from command line first, file second, and
	 * benchmark if not available.
	 */
	if (!loops_per_ms) {
		if (benchmark)
			goto bench;
		if ((fp = fopen(fname, "r"))) {
			fscanf(fp, "%lu", &loops_per_ms);
			fclose(fp);
			if (loops_per_ms) {
				printf("%lu loops_per_ms read from file interbench.loops_per_ms\n",
					loops_per_ms);
				goto loops_known;
			}
		}
bench:
		printf("loops_per_ms unknown; benchmarking...\n");

		/*
		 * To get as accurate a loop as possible we time it running
		 * SCHED_FIFO if we can
		 */
		get_fifo();
		calibrate_loop();
		get_normal();
	} else
		printf("loops_per_ms specified from command line\n");

	if (!(fp = fopen(fname, "w"))) {
		printf("Unable to write to file interbench.loops_per_ms\n");
		goto loops_known;
	}
	fprintf(fp, "%lu", loops_per_ms);
	printf("%lu loops_per_ms saved to file interbench.loops_per_ms\n",
		loops_per_ms);
	fclose(fp);

loops_known:
	printf("\n");
	get_ram();
	setup_threads();

	for (i = 0 ; i < BENCH_THREADS ; i++) {
		struct thread *thi = threadlist[i];

		printf("--- Benchmarking %s ---\n", thi->label);
		printf("Load\tLatency +/- SD (ms)  Completed\t%% Deadlines Met\n");

		printf("None:\t");
		sleep(1);
		initialise_thread(thi);
		bench(thi);
		fill_table(i, i);
		show_latencies(i);

		for (j = 0 ; j < THREADS ; j++) {
			struct thread *thj = threadlist[j];

			if (j == i)
				continue;
			printf("%s:\t", thj->label);
			sleep(1);
			initialise_thread(thi);
			initialise_thread(thj);
			start_thread(thj);
			bench(thi);
			stop_thread(thj);
			fill_table(i, j);
			show_latencies(i);
		}
		printf("\n");
	}

	return 0;
}
